from dataclasses import dataclass
from types import SimpleNamespace

import pytest

from mycli import clibuffer


@dataclass
class DummyDocument:
    text: str


@dataclass
class DummyBuffer:
    document: DummyDocument


@dataclass
class DummyLayout:
    buffer: DummyBuffer
    requested_names: list[str]

    def get_buffer_by_name(self, name: str) -> DummyBuffer:
        self.requested_names.append(name)
        return self.buffer


def make_app_for_text(text: str) -> tuple[SimpleNamespace, DummyLayout]:
    layout = DummyLayout(
        buffer=DummyBuffer(document=DummyDocument(text=text)),
        requested_names=[],
    )
    return SimpleNamespace(layout=layout), layout


def test_multiline_exception_handles_favorite_queries_only_after_blank_line() -> None:
    assert clibuffer._multiline_exception(r'\fs demo select 1; select 2') is False
    assert clibuffer._multiline_exception('\\fs demo select 1; select 2\n') is True


@pytest.mark.parametrize(
    ('text', 'expected'),
    (
        (r'\dt', True),
        ('select 1 //', True),
        ('select 1 \\g', True),
        ('select 1 \\G', True),
        ('select 1 \\e', True),
        ('select 1 \\edit', True),
        ('select 1 \\clip', True),
        ('help topic', True),
        ('HELP topic', True),
        ('   ', True),
        ('select 1', False),
    ),
)
def test_multiline_exception_detects_commands_terminators_and_plain_sql(
    monkeypatch,
    text: str,
    expected: bool,
) -> None:
    monkeypatch.setattr(clibuffer.iocommands, 'get_current_delimiter', lambda: '//')
    monkeypatch.setattr(clibuffer, 'SPECIAL_COMMANDS', {'help': object(), 'exit': object()})

    assert clibuffer._multiline_exception(text) is expected


def test_cli_is_multiline_returns_false_when_multiline_mode_is_disabled(monkeypatch) -> None:
    mycli = SimpleNamespace(multi_line=False)

    def fail_get_app() -> None:
        raise AssertionError('get_app() should not be called when multiline mode is disabled')

    monkeypatch.setattr(clibuffer, 'get_app', fail_get_app)

    multiline_filter = clibuffer.cli_is_multiline(mycli)

    assert multiline_filter() is False


@pytest.mark.parametrize('text', ('help\tselect', 'HELP\nselect'))
def test_multiline_exception_recognizes_non_backslashed_special_commands_with_general_whitespace(
    monkeypatch,
    text: str,
) -> None:
    monkeypatch.setattr(clibuffer.iocommands, 'get_current_delimiter', lambda: ';')
    monkeypatch.setattr(clibuffer, 'SPECIAL_COMMANDS', {'help': object(), 'exit': object()})

    assert clibuffer._multiline_exception(text) is True


@pytest.mark.parametrize(
    ('text', 'expected'),
    (
        ('select 1', True),
        ('help select', False),
    ),
)
def test_cli_is_multiline_uses_buffer_text_when_multiline_mode_is_enabled(
    monkeypatch,
    text: str,
    expected: bool,
) -> None:
    app, layout = make_app_for_text(text)
    mycli = SimpleNamespace(multi_line=True)

    monkeypatch.setattr(clibuffer, 'get_app', lambda: app)
    monkeypatch.setattr(clibuffer.iocommands, 'get_current_delimiter', lambda: ';')
    monkeypatch.setattr(clibuffer, 'SPECIAL_COMMANDS', {'help': object()})

    multiline_filter = clibuffer.cli_is_multiline(mycli)

    assert multiline_filter() is expected
    assert layout.requested_names == [clibuffer.DEFAULT_BUFFER]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, cast

import mycli.main_modes.list_ssh_config as list_ssh_config_mode


@dataclass
class DummyCliArgs:
    ssh_config_path: str = 'ssh_config'
    verbose: int = 0


class DummyMyCli:
    def __init__(self, config: Any) -> None:
        self.config = config
        self.verbosity = 0


class DummySSHConfig:
    def __init__(self, hostnames: list[str] | Exception, lookups: dict[str, dict[str, str]] | None = None) -> None:
        self.hostnames = hostnames
        self.lookups = lookups or {}

    def get_hostnames(self) -> list[str]:
        if isinstance(self.hostnames, Exception):
            raise self.hostnames
        return self.hostnames

    def lookup(self, hostname: str) -> dict[str, str]:
        return self.lookups[hostname]


def main_list_ssh_config(cli_args: DummyCliArgs) -> int:
    mycli = DummyMyCli(config={})
    mycli.verbosity = cli_args.verbose
    return list_ssh_config_mode.main_list_ssh_config(cast(Any, mycli), cast(Any, cli_args))


def test_main_list_ssh_config_lists_hostnames(monkeypatch) -> None:
    secho_calls: list[tuple[str, bool | None, str | None]] = []
    ssh_config = DummySSHConfig(['prod', 'staging'])

    monkeypatch.setattr(list_ssh_config_mode, 'read_ssh_config', lambda _path: ssh_config)
    monkeypatch.setattr(
        list_ssh_config_mode.click,
        'secho',
        lambda message, err=None, fg=None: secho_calls.append((message, err, fg)),
    )

    result = main_list_ssh_config(DummyCliArgs(verbose=0))

    assert result == 0
    assert secho_calls == [
        ('prod', None, None),
        ('staging', None, None),
    ]


def test_main_list_ssh_config_lists_verbose_host_details(monkeypatch) -> None:
    secho_calls: list[tuple[str, bool | None, str | None]] = []
    ssh_config = DummySSHConfig(
        ['prod'],
        lookups={'prod': {'hostname': 'db.example.com'}},
    )

    monkeypatch.setattr(list_ssh_config_mode, 'read_ssh_config', lambda _path: ssh_config)
    monkeypatch.setattr(
        list_ssh_config_mode.click,
        'secho',
        lambda message, err=None, fg=None: secho_calls.append((message, err, fg)),
    )

    result = main_list_ssh_config(DummyCliArgs(verbose=1))

    assert result == 0
    assert secho_calls == [('prod : db.example.com', None, None)]


def test_main_list_ssh_config_reports_host_lookup_errors(monkeypatch) -> None:
    secho_calls: list[tuple[str, bool | None, str | None]] = []
    ssh_config = DummySSHConfig(KeyError('bad ssh config'))

    monkeypatch.setattr(list_ssh_config_mode, 'read_ssh_config', lambda _path: ssh_config)
    monkeypatch.setattr(
        list_ssh_config_mode.click,
        'secho',
        lambda message, err=None, fg=None: secho_calls.append((message, err, fg)),
    )

    result = main_list_ssh_config(DummyCliArgs())

    assert result == 1
    assert secho_calls == [('Error reading ssh config', True, 'red')]



from __future__ import annotations

import os
import sys
import traceback
from typing import TYPE_CHECKING, Any

import click
import keyring
import pymysql
from pymysql.constants.CR import CR_SERVER_LOST
from pymysql.constants.ER import ACCESS_DENIED_ERROR, HANDSHAKE_ERROR

from mycli.compat import WIN
from mycli.config import str_to_bool
from mycli.constants import (
    DEFAULT_CHARSET,
    DEFAULT_HOST,
    DEFAULT_PORT,
    EMPTY_PASSWORD_FLAG_SENTINEL,
    ER_MUST_CHANGE_PASSWORD_LOGIN,
)
from mycli.packages.filepaths import guess_socket_location
from mycli.sqlexecute import SQLExecute

try:
    from pwd import getpwuid
except ImportError:
    pass


class ClientConnectionMixin:
    if TYPE_CHECKING:
        my_cnf: Any
        config: Any
        config_without_package_defaults: Any
        keepalive_ticks: int | None
        sandbox_mode: bool
        sqlexecute: Any
        logger: Any

        def read_my_cnf(self, files: Any, keys: list[str]) -> dict[str, Any]: ...
        def merge_ssl_with_cnf(self, ssl_config: dict[str, Any], cnf: dict[str, Any]) -> dict[str, Any] | None: ...
        def echo(self, *args: Any, **kwargs: Any) -> None: ...

    def connect(
        self,
        database: str | None = "",
        user: str | None = "",
        passwd: str | int | None = None,
        host: str | None = "",
        port: str | int | None = "",
        socket: str | None = "",
        character_set: str | None = "",
        local_infile: bool | None = False,
        ssl: dict[str, Any] | None = None,
        ssh_user: str | None = "",
        ssh_host: str | None = "",
        ssh_port: int = 22,
        ssh_password: str | None = "",
        ssh_key_filename: str | None = "",
        init_command: str | None = "",
        unbuffered: bool | None = None,
        use_keyring: bool | None = None,
        reset_keyring: bool | None = None,
        keepalive_ticks: int | None = None,
    ) -> None:
        cnf = {
            "database": None,
            "user": None,
            "password": None,
            "host": None,
            "port": None,
            "socket": None,
            "default_socket": None,
            "default-character-set": None,
            "local-infile": None,
            "loose-local-infile": None,
            "ssl-ca": None,
            "ssl-cert": None,
            "ssl-key": None,
            "ssl-cipher": None,
            "ssl-verify-server-cert": None,
        }

        cnf = self.read_my_cnf(self.my_cnf, list(cnf.keys()))

        # Fall back to config values only if user did not specify a value.
        database = database or cnf["database"]
        user = user or cnf["user"] or os.getenv("USER")
        host = host or cnf["host"]
        port = port or cnf["port"]
        ssl_config: dict[str, Any] = ssl or {}
        user_connection_config = self.config_without_package_defaults.get('connection', {})
        self.keepalive_ticks = keepalive_ticks

        int_port = port and int(port)
        if not int_port:
            int_port = DEFAULT_PORT
            if not host or host == DEFAULT_HOST:
                socket = (
                    socket
                    or user_connection_config.get("default_socket")
                    or cnf["socket"]
                    or cnf["default_socket"]
                    or guess_socket_location()
                )

        passwd = passwd if isinstance(passwd, (str, int)) else cnf["password"]

        # default_character_set doesn't check in self.config_without_package_defaults, because the
        # option already existed before the my.cnf deprecation.  For the same reason,
        # default_character_set can be in [connection] or [main].
        if not character_set:
            if 'default_character_set' in self.config['connection']:
                character_set = self.config['connection']['default_character_set']
            elif 'default_character_set' in self.config['main']:
                character_set = self.config['main']['default_character_set']
            elif 'default_character_set' in cnf:
                character_set = cnf['default_character_set']
            elif 'default-character-set' in cnf:
                character_set = cnf['default-character-set']
        if not character_set:
            character_set = DEFAULT_CHARSET

        # Favor whichever local_infile option is set.
        use_local_infile = False
        for local_infile_option in (
            local_infile,
            user_connection_config.get('default_local_infile'),
            cnf['local_infile'],
            cnf['local-infile'],
            cnf['loose_local_infile'],
            cnf['loose-local-infile'],
            False,
        ):
            try:
                use_local_infile = str_to_bool(local_infile_option or '')
                break
            except (TypeError, ValueError):
                pass

        # temporary my.cnf override mappings
        if 'default_ssl_ca' in user_connection_config:
            cnf['ssl-ca'] = user_connection_config.get('default_ssl_ca') or None
        if 'default_ssl_cert' in user_connection_config:
            cnf['ssl-cert'] = user_connection_config.get('default_ssl_cert') or None
        if 'default_ssl_key' in user_connection_config:
            cnf['ssl-key'] = user_connection_config.get('default_ssl_key') or None
        if 'default_ssl_cipher' in user_connection_config:
            cnf['ssl-cipher'] = user_connection_config.get('default_ssl_cipher') or None
        if 'default_ssl_verify_server_cert' in user_connection_config:
            cnf['ssl-verify-server-cert'] = user_connection_config.get('default_ssl_verify_server_cert') or None

        # todo: rewrite the merge method using self.config['connection'] instead of cnf, after removing my.cnf support
        ssl_config_or_none: dict[str, Any] | None = self.merge_ssl_with_cnf(ssl_config, cnf)

        # default_ssl_ca_path is not represented in my.cnf
        if 'default_ssl_ca_path' in self.config['connection'] and (not ssl_config_or_none or not ssl_config_or_none.get('capath')):
            if ssl_config_or_none is None:
                ssl_config_or_none = {}
            ssl_config_or_none['capath'] = self.config['connection']['default_ssl_ca_path'] or False

        # prune lone check_hostname=False
        if not any(v for v in ssl_config.values()):
            ssl_config_or_none = None

        # password hierarchy
        # 1. -p / --pass/--password CLI options
        # 2. --password-file CLI option
        # 3. envvar (MYSQL_PWD)
        # 4. DSN (mysql://user:password)
        # 5. cnf (.my.cnf / etc)
        # 6. keyring

        keyring_identifier = f'{user}@{host}:{"" if socket else int_port}:{socket or ""}'
        keyring_domain = 'mycli.net'
        keyring_retrieved_cleanly = False

        if passwd is None and use_keyring and not reset_keyring:
            passwd = keyring.get_password(keyring_domain, keyring_identifier)
            if passwd is not None:
                keyring_retrieved_cleanly = True

        # prompt for password if requested by user
        if passwd == EMPTY_PASSWORD_FLAG_SENTINEL:
            passwd = click.prompt(f"Enter password for {user}", hide_input=True, show_default=False, default='', type=str, err=True)
            keyring_retrieved_cleanly = False

        # should not fail, but will help the typechecker
        assert not isinstance(passwd, int)

        connection_info: dict[Any, Any] = {
            "database": database,
            "user": user,
            "password": passwd,
            "host": host,
            "port": int_port,
            "socket": socket,
            "character_set": character_set,
            "local_infile": use_local_infile,
            "ssl": ssl_config_or_none,
            "ssh_user": ssh_user,
            "ssh_host": ssh_host,
            "ssh_port": int(ssh_port) if ssh_port else None,
            "ssh_password": ssh_password,
            "ssh_key_filename": ssh_key_filename,
            "init_command": init_command,
            "unbuffered": unbuffered,
        }

        def _update_keyring(password: str | None, keyring_retrieved_cleanly: bool):
            if not password:
                return
            if reset_keyring or (use_keyring and not keyring_retrieved_cleanly):
                try:
                    saved_pw = keyring.get_password(keyring_domain, keyring_identifier)
                    if password != saved_pw or reset_keyring:
                        keyring.set_password(keyring_domain, keyring_identifier, password)
                        click.secho(f'Password saved to the system keyring at {keyring_domain}/{keyring_identifier}', err=True)
                except Exception as e:
                    click.secho(f'Password not saved to the system keyring: {e}', err=True, fg='red')

        def _connect(
            retry_ssl: bool = False,
            retry_password: bool = False,
            keyring_save_eligible: bool = True,
            keyring_retrieved_cleanly: bool = False,
        ) -> None:
            try:
                if keyring_save_eligible:
                    _update_keyring(connection_info["password"], keyring_retrieved_cleanly=keyring_retrieved_cleanly)
                self.sqlexecute = SQLExecute(**connection_info)
            except pymysql.OperationalError as e1:
                if e1.args[0] == HANDSHAKE_ERROR and ssl is not None and ssl.get("mode", None) == "auto":
                    # if we already tried and failed to connect without SSL, raise the error
                    if retry_ssl:
                        raise e1
                    # disable SSL and try to connect again
                    connection_info["ssl"] = None
                    _connect(
                        retry_ssl=True, keyring_retrieved_cleanly=keyring_retrieved_cleanly, keyring_save_eligible=keyring_save_eligible
                    )
                elif e1.args[0] == ACCESS_DENIED_ERROR and connection_info["password"] is None:
                    # if we already tried and failed to connect with a new password, raise the error
                    if retry_password:
                        raise e1
                    # ask the user for a new password and try to connect again
                    new_password = click.prompt(
                        f"Enter password for {user}", hide_input=True, show_default=False, default='', type=str, err=True
                    )
                    connection_info["password"] = new_password
                    keyring_retrieved_cleanly = False
                    _connect(
                        retry_password=True,
                        keyring_retrieved_cleanly=keyring_retrieved_cleanly,
                        keyring_save_eligible=keyring_save_eligible,
                    )
                elif e1.args[0] == ER_MUST_CHANGE_PASSWORD_LOGIN:
                    self.echo(
                        "Your password has expired and the server rejected the connection.",
                        err=True,
                        fg='red',
                    )
                    raise e1
                elif e1.args[0] == CR_SERVER_LOST:
                    self.echo(
                        (
                            "Connection to server lost. If this error persists, it may be a mismatch between the server and "
                            "client SSL configuration. To troubleshoot the issue, try --ssl-mode=off or --ssl-mode=on."
                        ),
                        err=True,
                        fg='red',
                    )
                    raise e1
                else:
                    raise e1

        try:
            if not WIN and socket:
                try:
                    socket_owner = getpwuid(os.stat(socket).st_uid).pw_name
                except KeyError:
                    socket_owner = '<unknown>'
                self.echo(f"Connecting to socket {socket}, owned by user {socket_owner}", err=True)
                try:
                    _connect(keyring_retrieved_cleanly=keyring_retrieved_cleanly)
                except pymysql.OperationalError as e:
                    # These are "Can't open socket" and 2x "Can't connect"
                    if [code for code in (2001, 2002, 2003) if code == e.args[0]]:
                        self.logger.debug("Database connection failed: %r.", e)
                        self.logger.error("traceback: %r", traceback.format_exc())
                        self.logger.debug("Retrying over TCP/IP")
                        self.echo(f"Failed to connect to local MySQL server through socket '{socket}':")
                        self.echo(str(e), err=True)
                        self.echo("Retrying over TCP/IP", err=True)

                        # Else fall back to TCP/IP localhost
                        socket = ""
                        host = DEFAULT_HOST
                        port = DEFAULT_PORT
                        # todo should reload the keyring identifier here instead of invalidating
                        _connect(keyring_save_eligible=False)
                    else:
                        raise e
            else:
                host = host or DEFAULT_HOST
                port = port or DEFAULT_PORT
                # could try loading the keyring again here instead of assuming nothing important changed

                # Bad ports give particularly daft error messages
                try:
                    port = int(port)
                except ValueError:
                    self.echo(f"Error: Invalid port number: '{port}'.", err=True, fg="red")
                    sys.exit(1)

                _connect(keyring_retrieved_cleanly=keyring_retrieved_cleanly)

            # Check if SQLExecute detected sandbox mode during connection
            if self.sqlexecute and self.sqlexecute.sandbox_mode:
                self.sandbox_mode = True
                self.echo(
                    "Your password has expired. Use ALTER USER or SET PASSWSORD to set a new password, or quit.",
                    err=True,
                    fg='yellow',
                )
        except Exception as e:  # Connecting to a database could fail.
            self.logger.debug("Database connection failed: %r.", e)
            self.logger.error("traceback: %r", traceback.format_exc())
            self.echo(str(e), err=True, fg="red")
            sys.exit(1)

    def reconnect(self, database: str = "") -> bool:
        """
        Attempt to reconnect to the server. Return True if successful,
        False if unsuccessful.

        The "database" argument is used only to improve messages.
        """
        assert self.sqlexecute is not None
        assert self.sqlexecute.conn is not None

        # First pass with ping(reconnect=False) and minimal feedback levels.  This definitely
        # works as expected, and is a good idea especially when "connect" was used as a
        # synonym for "use".
        try:
            self.sqlexecute.conn.ping(reconnect=False)
            if not database:
                self.echo("Already connected.", fg="yellow")
            return True
        except pymysql.err.Error:
            pass

        # Second pass with ping(reconnect=True).  It is not demonstrated that this pass ever
        # gives the benefit it is looking for, _ie_ preserves session state.  We need to test
        # this with connection pooling.
        try:
            old_connection_id = self.sqlexecute.connection_id
            self.logger.debug("Attempting to reconnect.")
            self.echo("Reconnecting...", fg="yellow")
            self.sqlexecute.conn.ping(reconnect=True)
            # if a database is currently selected, set it on the conn again
            if self.sqlexecute.dbname:
                self.sqlexecute.conn.select_db(self.sqlexecute.dbname)
            self.logger.debug("Reconnected successfully.")
            self.echo("Reconnected successfully.", fg="yellow")
            self.sqlexecute.reset_connection_id()
            if old_connection_id != self.sqlexecute.connection_id:
                self.echo("Any session state was reset.", fg="red")
            return True
        except pymysql.err.Error:
            pass

        # Third pass with sqlexecute.connect() should always work, but always resets session state.
        try:
            self.logger.debug("Creating new connection")
            self.echo("Creating new connection...", fg="yellow")
            self.sqlexecute.connect()
            self.logger.debug("New connection created successfully.")
            self.echo("New connection created successfully.", fg="yellow")
            self.echo("Any session state was reset.", fg="red")
            return True
        except pymysql.OperationalError as e:
            self.logger.debug("Reconnect failed. e: %r", e)
            self.echo(str(e), err=True, fg="red")
            return False



# type: ignore

import pytest

import mycli.sqlexecute
from test.utils import CHARACTER_SET, DATABASE, HOST, PASSWORD, PORT, SSH_HOST, SSH_PORT, SSH_USER, USER, create_db, db_connection


@pytest.fixture(scope="function")
def connection():
    create_db(DATABASE)
    connection = db_connection(DATABASE)
    yield connection

    connection.close()


@pytest.fixture
def cursor(connection):
    with connection.cursor() as cur:
        return cur


@pytest.fixture
def executor(connection):
    return mycli.sqlexecute.SQLExecute(
        database=DATABASE,
        user=USER,
        host=HOST,
        password=PASSWORD,
        port=PORT,
        socket=None,
        character_set=CHARACTER_SET,
        local_infile=False,
        ssl=None,
        ssh_user=SSH_USER,
        ssh_host=SSH_HOST,
        ssh_port=SSH_PORT,
        ssh_password=None,
        ssh_key_filename=None,
    )



from prompt_toolkit.formatted_text import FormattedText

from mycli.packages.sqlresult import SQLResult


def test_sqlresult_str_includes_all_fields() -> None:
    result = SQLResult(
        preamble='before',
        header=['id'],
        rows=[(1,)],
        postamble='after',
        status='ok',
        command={'name': 'watch', 'seconds': 1.0},
    )

    assert 'before' in str(result)
    assert "['id']" in str(result)
    assert '[(1,)]' in str(result)
    assert 'after' in str(result)
    assert 'ok' in str(result)
    assert "{'name': 'watch', 'seconds': 1.0}" in str(result)


def test_sqlresult_status_plain_handles_none_and_formatted_text() -> None:
    empty = SQLResult()
    formatted = SQLResult(status=FormattedText([('', '1 row in set'), ('', ', '), ('class:warn', '1 warning')]))

    assert empty.status_plain is None
    assert formatted.status_plain == '1 row in set, 1 warning'



# type: ignore

"""Steps for behavioral style tests are defined in this module.

Each step is defined by the string decorating it. This string is used
to call the step in "*.feature" file.

"""

from behave import then, when
import wrappers


@when("we save a named query")
def step_save_named_query(context):
    """Send \fs command."""
    context.cli.sendline("\\fs foo SELECT 12345")


@when("we use a named query")
def step_use_named_query(context):
    """Send \f command."""
    context.cli.sendline("\\f foo")


@when("we delete a named query")
def step_delete_named_query(context):
    """Send \fd command."""
    context.cli.sendline("\\fd foo")


@then("we see the named query saved")
def step_see_named_query_saved(context):
    """Wait to see query saved."""
    wrappers.expect_exact(context, "Saved.", timeout=2)


@then("we see the named query executed")
def step_see_named_query_executed(context):
    """Wait to see select output."""
    wrappers.expect_exact(context, "SELECT 12345", timeout=2)


@then("we see the named query deleted")
def step_see_named_query_deleted(context):
    """Wait to see query deleted."""
    wrappers.expect_exact(context, "foo: Deleted", timeout=2)


@when("we save a named query with parameters")
def step_save_named_query_with_parameters(context):
    """Send \fs command for query with parameters."""
    context.cli.sendline('\\fs foo_args SELECT $1, "$2", "$3"')


@when("we use named query with parameters")
def step_use_named_query_with_parameters(context):
    """Send \f command with parameters."""
    context.cli.sendline('\\f foo_args 101 second "third value"')


@then("we see the named query with parameters executed")
def step_see_named_query_with_parameters_executed(context):
    """Wait to see select output."""
    wrappers.expect_exact(context, 'SELECT 101, "second", "third value"', timeout=2)


@when("we use named query with too few parameters")
def step_use_named_query_with_too_few_parameters(context):
    """Send \f command with missing parameters."""
    context.cli.sendline("\\f foo_args 101")


@then("we see the named query with parameters fail with missing parameters")
def step_see_named_query_with_parameters_fail_with_missing_parameters(context):
    """Wait to see select output."""
    wrappers.expect_exact(context, "missing substitution for $2 in query:", timeout=2)


@when("we use named query with too many parameters")
def step_use_named_query_with_too_many_parameters(context):
    """Send \f command with extra parameters."""
    context.cli.sendline("\\f foo_args 101 102 103 104")


@then("we see the named query with parameters fail with extra parameters")
def step_see_named_query_with_parameters_fail_with_extra_parameters(context):
    """Wait to see select output."""
    wrappers.expect_exact(context, "query does not have substitution parameter $4:", timeout=2)



import os

from mycli.packages.special.dbcommands import (
    list_databases,
    list_tables,
    status,
)
from mycli.packages.special.iocommands import (
    clip_command,
    close_tee,
    copy_query_to_clipboard,
    disable_pager,
    disable_show_warnings,
    editor_command,
    enable_show_warnings,
    flush_pipe_once_if_written,
    forced_horizontal,
    get_clip_query,
    get_current_delimiter,
    get_editor_query,
    get_filename,
    is_expanded_output,
    is_pager_enabled,
    is_redirected,
    is_show_favorite_query,
    is_show_warnings_enabled,
    is_timing_enabled,
    open_external_editor,
    set_delimiter,
    set_destructive_keywords,
    set_expanded_output,
    set_favorite_queries,
    set_forced_horizontal_output,
    set_pager,
    set_pager_enabled,
    set_redirect,
    set_show_favorite_query,
    set_show_warnings_enabled,
    set_timing_enabled,
    split_queries,
    unset_once_if_written,
    write_once,
    write_pipe_once,
    write_tee,
)

if not os.environ.get('MYCLI_LLM_OFF'):
    from mycli.packages.special.llm import (
        FinishIteration,
        handle_llm,
        is_llm_command,
        sql_using_llm,
    )
else:

    class FinishIteration(Exception):  # type: ignore[no-redef]
        def __init__(self, results=None):
            self.results = results

    def is_llm_command(command: str) -> bool:  # type: ignore[no-redef]
        return False

    def handle_llm(*args, **kwargs):  # type: ignore[no-redef, misc]
        raise FinishIteration(results=None)

    def sql_using_llm(*args, **kwargs):  # type: ignore[no-redef, misc]
        raise FinishIteration(results=None)


from mycli.packages.special.main import (
    CommandNotFound,
    SpecialCommandAlias,
    execute,
    parse_special_command,
    register_special_command,
    special_command,
)

__all__: list[str] = [
    'CommandNotFound',
    'FinishIteration',
    'SpecialCommandAlias',
    'clip_command',
    'close_tee',
    'copy_query_to_clipboard',
    'disable_pager',
    'disable_show_warnings',
    'editor_command',
    'enable_show_warnings',
    'execute',
    'flush_pipe_once_if_written',
    'forced_horizontal',
    'get_clip_query',
    'get_current_delimiter',
    'get_editor_query',
    'get_filename',
    'handle_llm',
    'is_expanded_output',
    'is_llm_command',
    'is_pager_enabled',
    'is_redirected',
    'is_show_warnings_enabled',
    'is_timing_enabled',
    'list_databases',
    'list_tables',
    'open_external_editor',
    'parse_special_command',
    'register_special_command',
    'set_delimiter',
    'set_destructive_keywords',
    'set_expanded_output',
    'set_favorite_queries',
    'set_forced_horizontal_output',
    'set_pager',
    'set_pager_enabled',
    'set_redirect',
    'set_show_warnings_enabled',
    'set_timing_enabled',
    'set_show_favorite_query',
    'is_show_favorite_query',
    'special_command',
    'split_queries',
    'sql_using_llm',
    'status',
    'unset_once_if_written',
    'write_once',
    'write_pipe_once',
    'write_tee',
]



# type: ignore

"""Steps for behavioral style tests are defined in this module.

Each step is defined by the string decorating it. This string is used
to call the step in "*.feature" file.

"""

from behave import then, when
import pexpect
import wrappers

from mycli.constants import DEFAULT_DATABASE


@when("we create database")
def step_db_create(context):
    """Send create database."""
    context.cli.sendline(f"create database {context.conf['dbname_tmp']};")

    context.response = {"database_name": context.conf["dbname_tmp"]}


@when("we drop database")
def step_db_drop(context):
    """Send drop database."""
    context.cli.sendline(f"drop database {context.conf['dbname_tmp']};")


@when("we connect to test database")
def step_db_connect_test(context):
    """Send connect to database."""
    db_name = context.conf["dbname"]
    context.currentdb = db_name
    context.cli.sendline(f"use {db_name};")


@when("we connect to quoted test database")
def step_db_connect_quoted_tmp(context):
    """Send connect to database."""
    db_name = context.conf["dbname"]
    context.currentdb = db_name
    context.cli.sendline(f"use `{db_name}`;")


@when("we connect to tmp database")
def step_db_connect_tmp(context):
    """Send connect to database."""
    db_name = context.conf["dbname_tmp"]
    context.currentdb = db_name
    context.cli.sendline(f"use {db_name}")


@when("we connect to dbserver")
def step_db_connect_dbserver(context):
    """Send connect to database."""
    context.currentdb = DEFAULT_DATABASE
    context.cli.sendline(f"use {DEFAULT_DATABASE}")


@then("dbcli exits")
def step_wait_exit(context):
    """Make sure the cli exits."""
    wrappers.expect_exact(context, pexpect.EOF, timeout=5)


@then("we see dbcli prompt")
def step_see_prompt(context):
    """Wait to see the prompt."""
    user = context.conf["user"]
    host = context.conf["host"]
    dbname = context.currentdb
    wrappers.wait_prompt(context, f"{user}@{host}:{dbname}> ")


@then("we see help output")
def step_see_help(context):
    for expected_line in context.fixture_data["help_commands.txt"]:
        # in case tests are run without extras
        if 'LLM' in expected_line:
            continue
        wrappers.expect_exact(context, expected_line, timeout=1)


@then("we see database created")
def step_see_db_created(context):
    """Wait to see create database output."""
    wrappers.expect_exact(context, "Query OK, 1 row affected", timeout=2)


@then("we see database dropped")
def step_see_db_dropped(context):
    """Wait to see drop database output."""
    wrappers.expect_exact(context, "Query OK, 0 rows affected", timeout=2)


@then("we see database dropped and no default database")
def step_see_db_dropped_no_default(context):
    """Wait to see drop database output."""
    user = context.conf["user"]
    host = context.conf["host"]
    database = "(none)"
    context.currentdb = None

    wrappers.expect_exact(context, "Query OK, 0 rows affected", timeout=2)
    wrappers.wait_prompt(context, f"{user}@{host}:{database}>")


@then("we see database connected")
def step_see_db_connected(context):
    """Wait to see drop database output."""
    wrappers.expect_exact(context, 'connected to database "', timeout=2)
    wrappers.expect_exact(context, '"', timeout=2)
    wrappers.expect_exact(context, f' as user "{context.conf["user"]}"', timeout=2)



from __future__ import annotations


class FavoriteQueries:
    section_name: str = "favorite_queries"

    usage = """
Favorite Queries are a way to save frequently used queries
with a short name.
Examples:

    # Save a new favorite query.
    > \\fs simple select * from abc where a is not Null;

    # List all favorite queries.
    > \\f
    ╒════════╤═══════════════════════════════════════╕
    │ Name   │ Query                                 │
    ╞════════╪═══════════════════════════════════════╡
    │ simple │ SELECT * FROM abc where a is not NULL │
    ╘════════╧═══════════════════════════════════════╛

    # Run a favorite query.
    > \\f simple
    ╒════════╤════════╕
    │ a      │ b      │
    ╞════════╪════════╡
    │ 日本語  │ 日本語  │
    ╘════════╧════════╛

    # Delete a favorite query.
    > \\fd simple
    simple: Deleted.
"""

    # Class-level variable, for convenience to use as a singleton.
    instance: FavoriteQueries

    def __init__(self, config) -> None:
        self.config = config

    @classmethod
    def from_config(cls, config):
        return FavoriteQueries(config)

    def list(self) -> list[str | None]:
        return list(self.config.get(self.section_name, {}))

    def get(self, name) -> str | None:
        return self.config.get(self.section_name, {}).get(name, None)

    def save(self, name: str, query: str) -> None:
        self.config.encoding = "utf-8"
        if self.section_name not in self.config:
            self.config[self.section_name] = {}
        self.config[self.section_name][name] = query
        self.config.write()

    def delete(self, name: str) -> str:
        try:
            del self.config[self.section_name][name]
        except KeyError:
            return f'{name}: Not Found.'
        self.config.write()
        return f'{name}: Deleted.'



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from mycli.client import MyCli


def main_list_dsn(mycli: 'MyCli') -> int:
    try:
        alias_dsn = mycli.config['alias_dsn']
    except KeyError:
        click.secho('Invalid DSNs found in the config file. Please check the "[alias_dsn]" section in myclirc.', err=True, fg='red')
        return 1
    except Exception as e:
        click.secho(str(e), err=True, fg='red')
        return 1
    for alias, value in alias_dsn.items():
        if mycli.verbosity >= 1:
            click.secho(f'{alias} : {value}')
        else:
            click.secho(alias)
    return 0



import logging
from typing import cast

from prompt_toolkit.styles import Style, merge_styles
from prompt_toolkit.styles.pygments import style_from_pygments_cls
from prompt_toolkit.styles.style import _MergedStyle
from pygments.style import Style as PygmentsStyle
import pygments.styles
from pygments.token import Token, string_to_tokentype
from pygments.util import ClassNotFound

logger = logging.getLogger(__name__)

# map Pygments tokens (ptk 1.0) to class names (ptk 2.0).
TOKEN_TO_PROMPT_STYLE: dict[Token, str] = {
    Token.Menu.Completions.Completion.Current: "completion-menu.completion.current",
    Token.Menu.Completions.Completion: "completion-menu.completion",
    Token.Menu.Completions.Meta.Current: "completion-menu.meta.completion.current",
    Token.Menu.Completions.Meta: "completion-menu.meta.completion",
    Token.Menu.Completions.MultiColumnMeta: "completion-menu.multi-column-meta",
    Token.Menu.Completions.ProgressButton: "scrollbar.arrow",  # best guess
    Token.Menu.Completions.ProgressBar: "scrollbar",  # best guess
    Token.SelectedText: "selected",
    Token.SearchMatch: "search",
    Token.SearchMatch.Current: "search.current",
    Token.MatchingBracket.Cursor: "matching-bracket.cursor",
    Token.MatchingBracket.Other: "matching-bracket.other",
    Token.Toolbar: "bottom-toolbar",
    Token.Toolbar.Off: "bottom-toolbar.off",
    Token.Toolbar.On: "bottom-toolbar.on",
    Token.Toolbar.Search: "search-toolbar",
    Token.Toolbar.Search.Text: "search-toolbar.text",
    Token.Toolbar.System: "system-toolbar",
    Token.Toolbar.Arg: "arg-toolbar",
    Token.Toolbar.Arg.Text: "arg-toolbar.text",
    Token.Toolbar.Transaction.Valid: "bottom-toolbar.transaction.valid",
    Token.Toolbar.Transaction.Failed: "bottom-toolbar.transaction.failed",
    Token.Output.TableSeparator: "output.table-separator",
    Token.Output.Header: "output.header",
    Token.Output.OddRow: "output.odd-row",
    Token.Output.EvenRow: "output.even-row",
    Token.Output.Null: "output.null",
    Token.Output.Status: "output.status",
    Token.Output.Status.WarningCount: "output.status.warning-count",
    Token.Output.Timing: "output.timing",
    Token.Warnings.TableSeparator: "warnings.table-separator",
    Token.Warnings.Header: "warnings.header",
    Token.Warnings.OddRow: "warnings.odd-row",
    Token.Warnings.EvenRow: "warnings.even-row",
    Token.Warnings.Null: "warnings.null",
    Token.Warnings.Status: "warnings.status",
    Token.Warnings.Status.WarningCount: "warnings.status.warning-count",
    Token.Warnings.Timing: "warnings.timing",
    Token.Prompt: "prompt",
    Token.Continuation: "continuation",
}

# reverse dict for cli_helpers, because they still expect Pygments tokens.
PROMPT_STYLE_TO_TOKEN: dict[str, Token] = {v: k for k, v in TOKEN_TO_PROMPT_STYLE.items()}

# all tokens that the Pygments MySQL lexer can produce
OVERRIDE_STYLE_TO_TOKEN: dict[str, Token] = {
    "sql.comment": Token.Comment,
    "sql.comment.multi-line": Token.Comment.Multiline,
    "sql.comment.single-line": Token.Comment.Single,
    "sql.comment.optimizer-hint": Token.Comment.Special,
    "sql.escape": Token.Error,
    "sql.keyword": Token.Keyword,
    "sql.datatype": Token.Keyword.Type,
    "sql.literal": Token.Literal,
    "sql.literal.date": Token.Literal.Date,
    "sql.symbol": Token.Name,
    "sql.quoted-schema-object": Token.Name.Quoted,
    "sql.quoted-schema-object.escape": Token.Name.Quoted.Escape,
    "sql.constant": Token.Name.Constant,
    "sql.function": Token.Name.Function,
    "sql.variable": Token.Name.Variable,
    "sql.number": Token.Number,
    "sql.number.binary": Token.Number.Bin,
    "sql.number.float": Token.Number.Float,
    "sql.number.hex": Token.Number.Hex,
    "sql.number.integer": Token.Number.Integer,
    "sql.operator": Token.Operator,
    "sql.punctuation": Token.Punctuation,
    "sql.string": Token.String,
    "sql.string.double-quouted": Token.String.Double,
    "sql.string.escape": Token.String.Escape,
    "sql.string.single-quoted": Token.String.Single,
    "sql.whitespace": Token.Text,
}


def parse_pygments_style(
    token_name: str,
    style_object: type[PygmentsStyle] | PygmentsStyle | dict[object, str] | str,
    style_dict: dict[str, str],
) -> tuple[Token, str]:
    """Parse token type and style string.

    :param token_name: str name of Pygments token. Example: "Token.String"
    :param style_object: pygments.style.Style instance to use as base
    :param style_dict: dict of token names and their styles, customized to this cli

    """
    token_type = string_to_tokentype(token_name)
    if isinstance(style_object, type) and issubclass(style_object, PygmentsStyle):
        # When a Pygments Style class is passed, use its "styles" mapping.
        other_token_type = string_to_tokentype(style_dict[token_name])
        style_class = cast(type[PygmentsStyle], style_object)
        return token_type, style_class.styles[other_token_type]
    elif isinstance(style_object, PygmentsStyle):
        other_token_type = string_to_tokentype(style_dict[token_name])
        return token_type, style_object.styles[other_token_type]
    else:
        return token_type, style_dict[token_name]


def is_valid_pygments(name: str) -> bool:
    try:

        class TestStyle(PygmentsStyle):
            default_style = ''
            styles = {Token.Default: name}

        return True
    except AssertionError:
        # can't emit error because some styles are valid pygments and not valid ptoolkit
        return False


def is_valid_ptoolkit(name: str) -> bool:
    try:
        _s = Style([("default", name)])
        return True
    except ValueError:
        # can't emit error because some styles are valid pygments and not valid ptoolkit
        return False


def style_factory_ptoolkit(name: str, cli_style: dict[str, str]) -> _MergedStyle:
    try:
        style: PygmentsStyle = pygments.styles.get_style_by_name(name)
    except ClassNotFound:
        style = pygments.styles.get_style_by_name("native")

    prompt_styles: list[tuple[str, str]] = []
    # prompt-toolkit used pygments tokens for styling before, switched to style
    # names in 2.0. Convert old token types to new style names, for backwards compatibility.
    for token in cli_style:
        if token.startswith("Token."):
            # treat as pygments token (1.0)
            token_type, style_value = parse_pygments_style(token, style, cli_style)
            if token_type in TOKEN_TO_PROMPT_STYLE:
                prompt_style = TOKEN_TO_PROMPT_STYLE[token_type]
                if is_valid_ptoolkit(style_value):
                    prompt_styles.append((prompt_style, style_value))
            else:
                # we don't want to support tokens anymore
                logger.error("Unhandled style / class name: %s", token)
        else:
            # treat as prompt style name (2.0). See default style names here:
            # https://github.com/jonathanslenders/python-prompt-toolkit/blob/master/prompt_toolkit/styles/defaults.py
            if is_valid_ptoolkit(cli_style[token]):
                prompt_styles.append((token, cli_style[token]))

    override_style: Style = Style([("bottom-toolbar", "noreverse")])
    return merge_styles([style_from_pygments_cls(style), override_style, Style(prompt_styles)])


def style_factory_helpers(
    name: str,
    cli_style: dict[str, str],
    warnings: bool = False,
) -> PygmentsStyle:
    try:
        style: dict[PygmentsStyle | str, str] = pygments.styles.get_style_by_name(name).styles
    except ClassNotFound:
        style = pygments.styles.get_style_by_name("native").styles

    for token in cli_style:
        if token.startswith("Token."):
            token_type, style_value = parse_pygments_style(token, style, cli_style)
            if is_valid_pygments(style_value):
                style.update({token_type: style_value})
        elif token in PROMPT_STYLE_TO_TOKEN:
            token_type = PROMPT_STYLE_TO_TOKEN[token]
            if is_valid_pygments(cli_style[token]):
                style.update({token_type: cli_style[token]})
        elif token in OVERRIDE_STYLE_TO_TOKEN:
            token_type = OVERRIDE_STYLE_TO_TOKEN[token]
            if is_valid_pygments(cli_style[token]):
                style.update({token_type: cli_style[token]})
        else:
            # TODO: cli helpers will have to switch to ptk.Style
            logger.error("Unhandled style / class name: %s", token)

    if warnings:
        for warning_token in list(style.keys()):
            if 'Warnings' not in str(warning_token):
                continue
            warning_str = str(warning_token)
            output_str = warning_str.replace('Warnings', 'Output')
            output_token = string_to_tokentype(output_str)
            style[output_token] = style[warning_token]

    class OutputStyle(PygmentsStyle):
        default_style = ""
        styles = style

    return OutputStyle



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from prompt_toolkit.shortcuts import PromptSession
import sqlglot

from mycli.packages import special
from mycli.sqlexecute import SQLExecute

if TYPE_CHECKING:
    from mycli.client import MyCli


def server_date(sqlexecute: SQLExecute, quoted: bool = False) -> str:
    server_date_str = sqlexecute.now().strftime('%Y-%m-%d')
    if quoted:
        return f"'{server_date_str}'"
    else:
        return server_date_str


def server_datetime(sqlexecute: SQLExecute, quoted: bool = False) -> str:
    server_datetime_str = sqlexecute.now().strftime('%Y-%m-%d %H:%M:%S')
    if quoted:
        return f"'{server_datetime_str}'"
    else:
        return server_datetime_str


# todo: maybe these handlers belong in a repl_handlers.py (which does not exist yet)
# \clip doesn't even have a keybinding
def handle_clip_command(mycli: 'MyCli', text: str) -> bool:
    r"""A clip command is any query that is prefixed or suffixed by a
    '\clip'.

    :param text: Document
    :return: Boolean

    """

    if special.clip_command(text):
        query = special.get_clip_query(text) or mycli.get_last_query()
        message = special.copy_query_to_clipboard(sql=query)
        if message:
            raise RuntimeError(message)
        return True
    return False


def handle_editor_command(
    mycli: 'MyCli',
    text: str,
    inputhook: Callable | None,
    loaded_message_fn: Callable,
) -> str:
    r"""Editor command is any query that is prefixed or suffixed by a '\e'.
    The reason for a while loop is because a user might edit a query
    multiple times. For eg:

    "select * from \e"<enter> to edit it in vim, then come
    back to the prompt with the edited query "select * from
    blah where q = 'abc'\e" to edit it again.
    :param text: Document
    :return: Document

    """

    while special.editor_command(text):
        filename = special.get_filename(text)
        query = special.get_editor_query(text) or mycli.get_last_query()
        sql, message = special.open_external_editor(filename=filename, sql=query)
        if message:
            # Something went wrong. Raise an exception and bail.
            raise RuntimeError(message)
        while True:
            try:
                assert isinstance(mycli.prompt_session, PromptSession)
                text = mycli.prompt_session.prompt(
                    default=sql,
                    inputhook=inputhook,
                    message=loaded_message_fn,
                )
                break
            except KeyboardInterrupt:
                sql = ""

        continue
    return text


def handle_prettify_binding(
    mycli: 'MyCli',
    text: str,
) -> str:
    if not text:
        return ''
    try:
        statements = sqlglot.parse(text, read='mysql')
    except Exception:
        statements = []
    if len(statements) == 1 and statements[0]:
        parse_succeeded = True
        pretty_text = statements[0].sql(pretty=True, pad=4, dialect='mysql')
    else:
        parse_succeeded = False
        pretty_text = text.rstrip(';')
        mycli.toolbar_error_message = 'Prettify failed to parse single statement'
    if pretty_text and parse_succeeded:
        pretty_text = pretty_text + ';'
    return pretty_text


def handle_unprettify_binding(
    mycli: 'MyCli',
    text: str,
) -> str:
    if not text:
        return ''
    try:
        statements = sqlglot.parse(text, read='mysql')
    except Exception:
        statements = []
    if len(statements) == 1 and statements[0]:
        parse_succeeded = True
        unpretty_text = statements[0].sql(pretty=False, dialect='mysql')
    else:
        parse_succeeded = False
        unpretty_text = text.rstrip(';')
        mycli.toolbar_error_message = 'Unprettify failed to parse single statement'
    if unpretty_text and parse_succeeded:
        unpretty_text = unpretty_text + ';'
    return unpretty_text
